CREATE TABLE pgcryptokey (
	key_id SERIAL PRIMARY KEY,
	name TEXT DEFAULT 'main',
	encrypted_data_key BYTEA NOT NULL,
	created TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP NOT NULL,
	superseded_by INTEGER
);

-- Allow only one null superseded_by row per name
CREATE UNIQUE INDEX i_pgcryptokey ON pgcryptokey (name) WHERE superseded_by IS NULL;

--
-- Create key
--

CREATE FUNCTION create_cryptokey(name TEXT, byte_len INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS INTEGER
AS $$
DECLARE
	seq_name TEXT;
	seq_val BIGINT;
BEGIN
	IF (password IS NULL OR length(password) = 0) THEN
	    RAISE EXCEPTION 'invalid password';
	END IF;
	IF (byte_len <= 0) THEN
	    RAISE EXCEPTION 'invalid length';
	END IF;

	SELECT pg_get_serial_sequence('pgcryptokey', 'key_id') INTO seq_name;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'sequence name not found';
	END IF;
	SELECT nextval(seq_name) INTO seq_val;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'sequence value not found';
	END IF;

	-- we use the default options
	INSERT INTO pgcryptokey VALUES (seq_val, name,
		pgp_sym_encrypt_bytea(gen_random_bytes(byte_len), password), DEFAULT, NULL);
	RETURN seq_val;
END
$$ LANGUAGE plpgsql;

--
-- Superced key
--

CREATE FUNCTION supersede_cryptokey(v_key_id INTEGER, byte_len INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS INTEGER
AS $$
DECLARE
	seq_name TEXT;
	seq_val BIGINT;
	v_name TEXT;
BEGIN
	IF (password IS NULL OR length(password) = 0) THEN
	    RAISE EXCEPTION 'invalid password';
	END IF;
	IF (byte_len <= 0) THEN
	    RAISE EXCEPTION 'invalid length';
	END IF;

	SELECT pg_get_serial_sequence('pgcryptokey', 'key_id') INTO seq_name;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'sequence name not found';
	END IF;
	SELECT nextval(seq_name) INTO seq_val;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'sequence value not found';
	END IF;

	SELECT name INTO v_name
	FROM pgcryptokey
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	-- must update old row first because of the index
	UPDATE pgcryptokey SET superseded_by = seq_val
	WHERE name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	INSERT INTO pgcryptokey VALUES (seq_val, v_name,
		pgp_sym_encrypt_bytea(gen_random_bytes(byte_len), password), DEFAULT, NULL);
	RETURN seq_val;
END
$$ LANGUAGE plpgsql;

CREATE FUNCTION supersede_cryptokey(v_name TEXT, byte_len INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS INTEGER
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;
	RETURN supersede_cryptokey(v_key_id, byte_len, password);
END
$$ LANGUAGE plpgsql;

--
-- Change password
--

CREATE FUNCTION change_cryptokey_password(v_key_id INTEGER, old_password TEXT, new_password TEXT)
RETURNS VOID
AS $$
BEGIN
	IF (new_password IS NULL OR length(new_password) = 0) THEN
	    RAISE EXCEPTION 'invalid new password';
	END IF;
	-- old password is checked here
	UPDATE pgcryptokey SET encrypted_data_key = pgp_sym_encrypt_bytea(
		pgp_sym_decrypt_bytea(pgcryptokey.encrypted_data_key, old_password),
		new_password)
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not updated';
	END IF;
END
$$ LANGUAGE plpgsql;

CREATE FUNCTION change_cryptokey_password(v_name TEXT, old_password TEXT, new_password TEXT)
RETURNS VOID
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;
	PERFORM change_cryptokey_password(v_key_id, old_password, new_password);
END
$$ LANGUAGE plpgsql;

--
-- Store key in server variable
--

CREATE FUNCTION set_cryptokey(v_key_id INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS VOID
AS $$
DECLARE
	bytea_output TEXT;
	key BYTEA;
	v_name TEXT;
BEGIN
	-- only output bytea into server variables as hex
	SELECT current_setting('bytea_output', false) INTO bytea_output;
	IF bytea_output != 'hex' THEN	
	    RAISE EXCEPTION 'bytea_output must be set to ''hex''';
	END IF;
	SELECT name, pgp_sym_decrypt_bytea(pgcryptokey.encrypted_data_key, password) INTO v_name, key
	FROM pgcryptokey
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	PERFORM set_config('pgcryptokey.key_id', v_key_id::text, false);
	PERFORM set_config('pgcryptokey.name', v_name, false);
	-- We can only store text in server variables, not bytea, and
	-- pgp_sym_* functions can only take text as a password, so we
	-- will use the text representation of the bytea value as the
	-- password.
	PERFORM set_config('pgcryptokey.key', key::text, false);
END
$$ LANGUAGE plpgsql;

CREATE FUNCTION set_cryptokey(v_name TEXT, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS VOID
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;
	PERFORM set_cryptokey(v_key_id, password);
END
$$ LANGUAGE plpgsql;

--
-- Get key
--

CREATE FUNCTION get_cryptokey(v_key_id INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS TEXT
AS $$
DECLARE
	bytea_output TEXT;
	key BYTEA;
BEGIN
	-- only output bytea into server variables as hex
	SELECT current_setting('bytea_output', false) INTO bytea_output;
	IF bytea_output != 'hex' THEN	
	    RAISE EXCEPTION 'bytea_output must be set to ''hex''';
	END IF;
	SELECT pgp_sym_decrypt_bytea(pgcryptokey.encrypted_data_key, password)::text INTO key
	FROM pgcryptokey
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;
	RETURN key;
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

CREATE FUNCTION get_cryptokey(v_name TEXT, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS TEXT
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;
	RETURN get_cryptokey(v_key_id, password);
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

--
-- Reencrypt data
--

-- we do not allow key names here since it could be confusing for superseded keys
CREATE FUNCTION reencrypt_data(data BYTEA, old_key_id INTEGER, new_key_id INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS BYTEA
AS $$
BEGIN
	RETURN pgp_sym_encrypt(
		pgp_sym_decrypt(data, get_cryptokey(old_key_id, password)),
		get_cryptokey(new_key_id, password));
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

CREATE FUNCTION reencrypt_data_bytea(data BYTEA, old_key_id INTEGER, new_key_id INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS BYTEA
AS $$
BEGIN
	RETURN pgp_sym_encrypt_bytea(
		pgp_sym_decrypt_bytea(data, get_cryptokey(old_key_id, password)),
		get_cryptokey(new_key_id, password));
END
$$ LANGUAGE plpgsql PARALLEL SAFE;

--
-- Drop key
--

CREATE FUNCTION drop_cryptokey(v_key_id INTEGER, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS VOID
AS $$
BEGIN
	-- verify the key password
	PERFORM set_cryptokey(v_key_id, password);
	DELETE FROM pgcryptokey
	WHERE pgcryptokey.key_id = v_key_id;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;
END
$$ LANGUAGE plpgsql;

-- this only deletes the active key, but warns if other matching names exist
CREATE FUNCTION drop_cryptokey(v_name TEXT, password TEXT DEFAULT current_setting('pgcryptokey.default_password', true))
RETURNS VOID
AS $$
DECLARE
	v_key_id INTEGER;
BEGIN
	-- notify if superseded keys exist
	PERFORM key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NOT NULL;
	IF FOUND THEN
	    RAISE NOTICE 'superseded cryptokey entries for this name exist';
	END IF;

	SELECT key_id INTO v_key_id
	FROM pgcryptokey
	WHERE pgcryptokey.name = v_name AND superseded_by IS NULL;
	IF NOT FOUND THEN
	    RAISE EXCEPTION 'cryptokey entry not found';
	END IF;

	PERFORM drop_cryptokey(v_key_id, password);
END
$$ LANGUAGE plpgsql;
